@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thamos
-# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
+# Copyright(C) 2018 - 2021 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,9 +25,10 @@ import subprocess
 import sys
 import typing
 from functools import wraps
-from typing import Tuple
+from typing import Dict
 from typing import Optional
 from typing import Set
+from typing import Tuple
 
 import yaml
 import click
@@ -44,14 +45,21 @@ from thamos.exceptions import NoProjectDirError
 from thamos.exceptions import NoRuntimeEnvironmentError
 from thamos.config import config as configuration
 from thamos.lib import advise_here as thoth_advise_here
-from thamos.lib import provenance_check as thoth_provenance_check
-from thamos.lib import list_hardware_environments
-from thamos.lib import list_python_package_indexes
-from thamos.lib import list_thoth_s2i
+from thamos.lib import add_requirements_to_project
+from thamos.lib import collect_support_information_dict
 from thamos.lib import get_log
-from thamos.lib import install as thamos_install
+from thamos.lib import get_package_from_imported_packages
 from thamos.lib import get_status
-from thamos.lib import write_configuration, write_files, load_files
+from thamos.lib import install as thamos_install
+from thamos.lib import list_python_package_indexes
+from thamos.lib import list_python_environments
+from thamos.lib import list_thoth_container_images
+from thamos.lib import load_files
+from thamos.lib import provenance_check as thoth_provenance_check
+from thamos.lib import print_dependency_graph
+from thamos.lib import get_verified_packages_from_static_analysis
+from thamos.lib import write_configuration
+from thamos.lib import write_files
 from thamos.utils import workdir
 from thamos import __version__ as thamos_version
 
@@ -94,7 +102,7 @@ def handle_cli_exception(func: typing.Callable) -> typing.Callable:
     return wrapper
 
 
-def _print_report(report: dict, json_output: bool = False, title: Optional[str] = None):
+def _print_report(report: list, json_output: bool = False, title: Optional[str] = None):
     """Print reasoning to user."""
     if json_output:
         click.echo(json.dumps(report, sort_keys=True, indent=2))
@@ -148,6 +156,32 @@ def _print_report(report: dict, json_output: bool = False, title: Optional[str]
     console.print(table, justify="center")
 
 
+def _parse_labels(label: Optional[str]) -> Optional[Dict[str, str]]:
+    """Parse labels from their string representation."""
+    if not label:
+        return None
+
+    labels: Dict[str, str] = {}
+    parts = label.split(",")
+    for part in parts:
+        lab = part.split("=")
+        if len(lab) != 2:
+            raise ValueError(f"Unknown label, expected key=val: {part}")
+
+        if lab[0] in labels:
+            raise ValueError(f"Duplicate label {lab[0]!r}")
+
+        if not lab[0]:
+            raise ValueError(f"Empty label key for label value {lab[1]!r}")
+
+        if not lab[1]:
+            raise ValueError(f"Empty label value for label key {lab[0]!r}")
+
+        labels[lab[0]] = lab[1]
+
+    return labels
+
+
 class AliasedGroup(click.Group):
     """Provide clever command aliases."""
 
@@ -183,6 +217,7 @@ class AliasedGroup(click.Group):
 @click.option(
     "--workdir",
     "-d",
+    metavar="DIR",
     type=str,
     default=None,
     help="Adjust working directory for sub-commands.",
@@ -190,6 +225,7 @@ class AliasedGroup(click.Group):
 @click.option(
     "--thoth-host",
     "-t",
+    metavar="HOST",
     type=str,
     default=None,
     help="Use selected host instead of the one stated in the configuration file.",
@@ -284,13 +320,16 @@ def _print_version(ctx, json_output: bool = False):
 )
 @click.argument("pip_args", nargs=-1)
 @handle_cli_exception
-def install(runtime_environment: str, dev: bool, pip_args: Tuple[str]) -> None:
+def install(
+    runtime_environment: str, dev: bool, pip_args: Tuple[str]
+) -> None:  # noqa: D412
     """Install dependencies as stated in the lock file.
 
     This command assumes requirements files are present and dependencies are already resolved.
     If that's not the case, issue `thamos advise` before running this command.
 
     Examples:
+
       thamos install --runtime-environment "testing"
 
       thamos install --dev
@@ -336,10 +375,13 @@ def _error_virtual_environment(virtualenv_path: str) -> None:
     help="Specify explicitly runtime environment to get recommendations for; "
     "defaults to the first entry in the configuration file.",
 )
-def run(command: typing.List[str], runtime_environment: Optional[str] = None) -> None:
+def run(
+    command: typing.List[str], runtime_environment: Optional[str] = None
+) -> None:  # noqa: D412
     """Run the command in virtual environment.
 
     Examples:
+
       thamos run ./app.py
 
       thamos run --runtime-environment "testing" -- flask --help
@@ -397,10 +439,13 @@ def venv(ctx, runtime_environment: Optional[str] = None) -> None:
     is_flag=True,
     help="Purge virtual environments for all the runtime environments configured.",
 )
-def purge(ctx, runtime_environment: Optional[str] = None, all: bool = False) -> None:
+def purge(
+    ctx, runtime_environment: Optional[str] = None, all: bool = False
+) -> None:  # noqa: D412
     """Remove virtual environment created.
 
     Examples:
+
       thamos purge
 
       thamos purge --runtime-environment "testing"
@@ -526,6 +571,16 @@ def purge(ctx, runtime_environment: Optional[str] = None, all: bool = False) ->
     show_default=True,
     help="Write advised manifest changes to a file.",
 )
+@click.option(
+    "--labels",
+    "-l",
+    envvar="THAMOS_LABELS",
+    type=str,
+    metavar="KEY1=VALUE1,KEY2=VALUE2",
+    default=None,
+    show_default=True,
+    help="Labels used used to label the request.",
+)
 @handle_cli_exception
 def advise(
     debug: bool = False,
@@ -539,8 +594,9 @@ def advise(
     dev: bool = False,
     no_user_stack: bool = False,
     install: bool = False,
+    labels: Optional[str] = None,
     write_advised_manifest_changes: Optional[str] = None,
-):
+) -> None:  # noqa: D412
     """Ask Thoth for recommendations on the application stack.
 
     Ask the remote Thoth service for advise on the application stack used. The command
@@ -549,7 +605,8 @@ def advise(
     remote service. Optionally, install packages resolved by Thoth.
 
     Examples:
-      thamos advise --runtime-environment "testing"
+
+      thamos advise --runtime-environment "testing" --labels foo=bar,qux=baz
 
       thamos advise --dev
 
@@ -564,59 +621,61 @@ def advise(
         _LOGGER.error("Cannot install dependencies if lock files are not written")
         sys.exit(1)
 
-    with cwd(configuration.get_overlays_directory(runtime_environment)):
-        if not dev and configuration.requirements_format == "pipenv":
-            _LOGGER.warning(
-                "Development dependencies will not be considered during the resolution process - see %s",
-                jl("no_dev"),
-            )
+    labels_dict = _parse_labels(labels)
 
-        # In CLI we always call to obtain only the best software stack (count is implicitly set to 1).
-        results = thoth_advise_here(
-            recommendation_type=recommendation_type,
-            src_path=configuration.config_path,
-            runtime_environment_name=runtime_environment,
-            debug=debug,
-            nowait=no_wait,
-            force=force,
-            source_type=ThothAdviserIntegrationEnum.CLI,
-            no_static_analysis=no_static_analysis,
-            dev=dev,
-            no_user_stack=no_user_stack,
-            verify_tls=configuration.tls_verify,
+    if not dev and configuration.requirements_format == "pipenv":
+        _LOGGER.warning(
+            "Development dependencies will not be considered during the resolution process - see %s",
+            jl("no_dev"),
         )
 
-        if not results:
-            return sys.exit(2)
+    # In CLI we always call to obtain only the best software stack (count is implicitly set to 1).
+    results = thoth_advise_here(
+        recommendation_type=recommendation_type,
+        src_path=configuration.config_path,
+        runtime_environment_name=runtime_environment,
+        debug=debug,
+        nowait=no_wait,
+        force=force,
+        source_type=ThothAdviserIntegrationEnum.CLI,
+        no_static_analysis=no_static_analysis,
+        dev=dev,
+        no_user_stack=no_user_stack,
+        verify_tls=configuration.tls_verify,
+        labels=labels_dict,
+    )
 
-        if no_wait:
-            # Echo the analysis id to user when not waiting.
-            click.echo(results)
-            sys.exit(0)
+    if not results:
+        sys.exit(2)
 
-        result, error = results
-        if error:
-            if json_output:
-                json.dump(result, sys.stdout, indent=2)
-            else:
-                stack_info = (result.get("report") or {}).get("stack_info")
-                if stack_info:
-                    _print_report(
-                        stack_info,
-                        json_output=json_output,
-                        title="Application stack guidance",
-                    )
+    if no_wait:
+        # Echo the analysis id to user when not waiting.
+        click.echo(results)
+        sys.exit(0)
 
-                Console().print(
-                    result.get("error_msg")
-                    or "No error message was provided by the service.",
-                    style="bold red",
-                    justify="center",
+    result, error = results
+    if error:
+        if json_output:
+            json.dump(result, sys.stdout, indent=2)
+        else:
+            stack_info = (result.get("report") or {}).get("stack_info")
+            if stack_info:
+                _print_report(
+                    stack_info,
+                    json_output=json_output,
+                    title="Application stack guidance",
                 )
 
-            sys.exit(4)
+            Console().print(
+                result.get("error_msg")
+                or "No error message was provided by the service.",
+                style="bold red",
+                justify="center",
+            )
 
-        if not no_write:
+        sys.exit(4)
+    if not no_write:
+        with cwd(configuration.get_overlays_directory(runtime_environment)):
             if result["report"] and result["report"]["stack_info"]:
                 _print_report(
                     result["report"]["stack_info"],
@@ -660,8 +719,8 @@ def advise(
 
             if install:
                 thamos_install(runtime_environment_name=runtime_environment, dev=dev)
-        else:
-            click.echo(json.dumps(result, indent=2))
+    else:
+        click.echo(json.dumps(result, indent=2))
 
     sys.exit(0)
 
@@ -705,13 +764,14 @@ def provenance_check(
     json_output: bool = False,
     force: bool = False,
     runtime_environment: typing.Optional[str] = None,
-):
+) -> None:  # noqa: D412
     """Check provenance of installed packages.
 
     Collect information about direct dependencies and dependencies stated in the lock file
     and send them to the remote service to verify their provenance.
 
     Examples:
+
       thamos provenance-check --runtime-environment "production"
     """
     with cwd(configuration.get_overlays_directory(runtime_environment)):
@@ -750,8 +810,6 @@ def provenance_check(
         if any(item.get("type") == "ERROR" for item in report):
             sys.exit(4)
 
-        return 0
-
 
 @cli.command("log")
 @click.pass_context
@@ -784,12 +842,13 @@ def log(analysis_id: typing.Optional[str] = None):
 @handle_cli_exception
 def status(
     analysis_id: typing.Optional[str] = None, output_format: typing.Optional[str] = None
-):
+) -> None:  # noqa: D412
     """Get status of an analysis.
 
     If ANALYSIS_ID is not provided, the last request is used by default.
 
     Examples:
+
       thamos status
 
       thamos status "adviser-940101080006-110c392feb7cf6da"
@@ -825,13 +884,46 @@ def status(
     click.echo(output)
 
 
+@cli.command("graph")
+@click.pass_context
+@click.argument("analysis_id", type=str, required=False, metavar="ANALYSIS_ID")
+@click.option(
+    "--fold/--unfold",
+    is_flag=True,
+    envvar="THAMOS_GRAPH_FOLD",
+    default=True,
+    type=bool,
+    help="Collapse repeating sub-graphs in the output.",
+)
+@handle_cli_exception
+def graph(
+    analysis_id: typing.Optional[str] = None,
+    fold: bool = True,
+) -> None:  # noqa: D412
+    """Show dependency graph of resolved dependencies.
+
+    If ANALYSIS_ID is not provided, the last request is used by default.
+
+    Examples:
+
+      thamos graph
+
+      thamos graph "adviser-940101080006-110c392feb7cf6da"
+    """
+    with workdir():
+        printed = print_dependency_graph(analysis_id, fold=fold)
+        if not printed:
+            sys.exit(1)
+
+
 @cli.command("list")
 @click.pass_context
 @handle_cli_exception
-def list_() -> None:
+def list_() -> None:  # noqa: D412
     """List available runtime environments configured.
 
     Examples:
+
       thamos list
     """
     with workdir(configuration.CONFIG_NAME):
@@ -865,10 +957,13 @@ def list_() -> None:
     help="Specify explicitly runtime environment to be shown.",
 )
 @handle_cli_exception
-def show(output_format: str, runtime_environment: Optional[str] = None) -> None:
+def show(
+    output_format: str, runtime_environment: Optional[str] = None
+) -> None:  # noqa: D412
     """Show details for configured runtime environments.
 
     Examples:
+
       thamos show --runtime-environment "development"
     """
     with workdir(configuration.CONFIG_NAME):
@@ -970,13 +1065,14 @@ def config(no_interactive: bool = False, template: str = None):
     help="Specify output format for the status report.",
 )
 @handle_cli_exception
-def check(runtime_environment: Optional[str], output_format: str) -> None:
+def check(runtime_environment: Optional[str], output_format: str) -> None:  # noqa: D412
     """Check the configuration file and runtime environment.
 
     Check the correctness of the configuration file and runtime environment configuration
     for the current host.
 
     Examples:
+
         thamos check --runtime-environment "production"
 
         thamos check --output-format yaml
@@ -1023,7 +1119,7 @@ def check(runtime_environment: Optional[str], output_format: str) -> None:
     sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
 
 
-@cli.command("s2i")
+@cli.command("images")
 @click.pass_context
 @click.option(
     "--output-format",
@@ -1033,13 +1129,14 @@ def check(runtime_environment: Optional[str], output_format: str) -> None:
     help="Specify output format for the status report.",
 )
 @handle_cli_exception
-def s2i(output_format: str) -> None:
-    """Check available Thoth Source-To-Image containers.
+def images(output_format: str) -> None:  # noqa: D412
+    """Check available Thoth container images.
 
     Examples:
-      thamos s2i --output-format json
+
+      thamos images --output-format json
     """
-    result = list_thoth_s2i()
+    result = list_thoth_container_images()
 
     if output_format == "yaml":
         yaml.safe_dump({"s2i": result}, sys.stdout)
@@ -1067,7 +1164,7 @@ def s2i(output_format: str) -> None:
             row = []
             for key in header_sorted:
                 if key == "info":
-                    image_name = item.get("thoth_s2i_image_name")
+                    image_name = item.get("thoth_image_name")
                     row.append(jl(image_name.rsplit("/", maxsplit=1)[-1]))
                     continue
 
@@ -1082,63 +1179,6 @@ def s2i(output_format: str) -> None:
     sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
 
 
-@cli.command("hw")
-@click.pass_context
-@click.option(
-    "--output-format",
-    "-o",
-    type=click.Choice(["json", "yaml", "table"]),
-    default="table",
-    help="Specify output format for the status report.",
-)
-@handle_cli_exception
-def hw(output_format: str) -> None:
-    """List available hardware information on backend.
-
-    List available hardware information on the backend for which resolver can give more detailed information.
-    This hardware configuration is not enforced and is find if it does not match the one available on the
-    client side.
-    """
-    result = list_hardware_environments()
-
-    if output_format == "yaml":
-        yaml.safe_dump({"hw": result}, sys.stdout)
-    elif output_format == "json":
-        json.dump({"hw": result}, sys.stdout, indent=2)
-        sys.stdout.write("\n")
-    elif output_format == "table":
-        table = Table()
-
-        header = set()
-        for item in result:
-            for key in item.keys():
-                header.add(key)
-
-        header_sorted = sorted(header)
-        for item in header_sorted:
-            table.add_column(
-                item.replace("_", " ")
-                .replace("cpu", "CPU")
-                .replace("gpu", "GPU")
-                .replace("ram", "RAM"),
-                style="cyan",
-                overflow="fold",
-            )
-
-        for item in result:
-            row = []
-            for key in header_sorted:
-                entry = item.get(key)
-                row.append(str(entry) if entry is not None else "-")
-
-            table.add_row(*row)
-
-        console = Console()
-        console.print(table, justify="center")
-
-    sys.exit(0)
-
-
 @cli.command("indexes")
 @click.pass_context
 @click.option(
@@ -1224,13 +1264,14 @@ def add(
     runtime_environment: typing.Optional[str],
     index_url: str,
     dev: bool,
-) -> None:
+) -> None:  # noqa: D412
     """Add one or multiple requirements to the project.
 
     Add one or multiple requirements to the direct dependency listing without actually installing them.
     The supplied requirement is specified using PEP-508 standard.
 
     Examples:
+
       thamos add flask
 
       thamos add tensorflow --runtime-environment "training"
@@ -1239,22 +1280,12 @@ def add(
 
       thamos add 'importlib-metadata; python_version < "3.8"'
     """
-    project = configuration.get_project(runtime_environment, missing_dir_ok=True)
-    for req in requirement:
-        _LOGGER.info(
-            "Adding %r to %s requirements of runtime environment %r",
-            req,
-            "development" if dev else "default",
-            project.runtime_environment.name,
-        )
-        project.pipfile.add_requirement(
-            req, is_dev=dev, index_url=index_url, force=True
-        )
-
-    _LOGGER.warning(
-        "Changes done might require triggering new advise to resolve dependencies"
+    add_requirements_to_project(
+        requirement=requirement,
+        runtime_environment=runtime_environment,
+        index_url=index_url,
+        dev=dev,
     )
-    configuration.save_project(project)
 
 
 @cli.command("remove")
@@ -1272,10 +1303,11 @@ def add(
 def remove(
     requirement: typing.List[str],
     runtime_environment: typing.Optional[str],
-) -> None:
+) -> None:  # noqa: D412
     """Remove the given requirement.
 
     Examples:
+
       thamos remove flask
 
       thamos remove pytest --runtime-environment "training"
@@ -1317,4 +1349,135 @@ def remove(
     configuration.save_project(project)
 
 
+@cli.command("support")
+@click.option(
+    "--output",
+    "-o",
+    metavar="FILE",
+    type=str,
+    help="Specify output file",
+    default="-",
+)
+def support(output: str) -> None:
+    """Collect information from the current environment to report an issue."""
+    info = collect_support_information_dict()
+
+    if output == "-":
+        json.dump(info, sys.stdout, sort_keys=True, indent=2)
+    else:
+        with open(output, "w") as output_file:
+            json.dump(info, output_file, sort_keys=True, indent=2)
+
+
+@cli.command("whatprovides")
+@click.pass_context
+@click.option(
+    "--output-format",
+    "-o",
+    type=click.Choice(["json", "yaml", "table"]),
+    default="table",
+    help="Specify output format for the status report.",
+)
+@click.argument("import_name", type=str, required=True)
+@handle_cli_exception
+def whatprovides(import_name: str, output_format: str) -> None:  # noqa: D412
+    """For a given import returns list of packages with matching modules.
+
+    Examples:
+
+      thamos whatprovides sklearn
+
+      thamos whatprovides sklearn.linear_model.*
+    """
+    _LOGGER.info("Returning information on package %r", import_name)
+    result = get_package_from_imported_packages(import_name)
+
+    if output_format == "yaml":
+        yaml.safe_dump({"packages": result}, sys.stdout)
+    elif output_format == "json":
+        json.dump({"packages": result}, sys.stdout, indent=2)
+        sys.stdout.write("\n")
+    else:
+        _print_report(
+            result,
+            json_output=False,
+            title=f"Packages for {import_name}",
+        )
+
+    sys.exit(0)
+
+
+@cli.command("discover")
+@click.pass_context
+@click.option(
+    "--src-path",
+    "-s",
+    type=str,
+    default=".",
+    envvar="THAMOS_SOURCE_PATH",
+    help="Specify path to consider to discover packages.",
+)
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment to which the given package should be added.",
+)
+@handle_cli_exception
+def discover(
+    runtime_environment: typing.Optional[str], src_path: str = "."
+) -> None:  # noqa: D412
+    """Discover packages used in the project or in the file.
+
+    Examples:
+
+      thamos discover
+    """
+    # Obtain list of imports using invectio and verify package from PyPI
+    verified_packages = get_verified_packages_from_static_analysis(src_path=src_path)
+
+    # Update requirements files (Pipfile/Pipfile.lock) or requirements.txt (requirements logic)
+    for package in verified_packages:
+        add_requirements_to_project(
+            requirement=[package["package_name"]],
+            runtime_environment=runtime_environment,
+            index_url=package["index_url"],
+            dev=False,
+        )
+
+
+@cli.command("environments")
+@click.option(
+    "--output-format",
+    "-o",
+    type=click.Choice(["json", "yaml", "table"]),
+    default="table",
+    help="Specify output format for the status report.",
+)
+@click.pass_context
+@handle_cli_exception
+def environments_(output_format: str) -> None:  # noqa: D412
+    """Show available Python environments.
+
+    Examples:
+
+      thamos environments
+    """
+    environments = list_python_environments()
+
+    if output_format == "yaml":
+        yaml.safe_dump({"environment": environments}, sys.stdout)
+    elif output_format == "json":
+        json.dump({"environment": environments}, sys.stdout, indent=2)
+        sys.stdout.write("\n")
+    else:
+        _print_report(
+            environments,
+            json_output=False,
+            title="Python environments",
+        )
+
+
 __name__ == "__main__" and cli()
