@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thamos
-# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
+# Copyright(C) 2018 - 2021 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -67,6 +67,7 @@ _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA = {
         "cudnn_version": {"type": ["string", "null"]},
         "mkl_version": {"type": ["string", "null"]},
         "base_image": {"type": ["string", "null"]},
+        "labels": {"type": ["object", "null"]},
         "hardware": {
             "type": "object",
             "properties": {
@@ -119,6 +120,7 @@ _CONFIG_SCHEMA = {
         "runtime_environments": {
             "type": "array",
             "items": _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA,
+            "minItems": 1,
         },
         "managers": {
             "type": "array",
@@ -759,7 +761,7 @@ class _Configuration:
         old_project = self.get_project(project.runtime_environment.name)
         if old_project.runtime_environment != project.runtime_environment:
             self.set_runtime_environment(
-                runtime_environment=project.runtime_environment
+                runtime_environment=project.runtime_environment.to_dict()
             )
             self.save_config()
 
