@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thamos
-# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
+# Copyright(C) 2018 - 2021 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,10 +17,12 @@
 
 """Core parts of library for interacting with Thoth."""
 
-import os
 import logging
-import typing
+import os
 import platform
+import sys
+import typing
+from itertools import chain
 from time import sleep
 from time import monotonic
 from contextlib import contextmanager
@@ -44,6 +46,7 @@ from thoth.common import ThothAdviserIntegrationEnum
 from thoth.common import get_justification_link as jl
 from thoth.common import cwd
 
+import thamos.discover
 from . import __version__ as thamos_version
 from .swagger_client.rest import ApiException
 from .swagger_client import ApiClient
@@ -55,8 +58,8 @@ from .swagger_client import ProvenanceInput
 from .swagger_client import AdviseApi
 from .swagger_client import ImageAnalysisApi
 from .swagger_client import ProvenanceApi
-from .swagger_client import HardwareApi
-from .swagger_client import S2iApi
+from .swagger_client import ContainerImagesApi
+from .swagger_client import EnvironmentsApi
 from .swagger_client import PythonPackagesApi
 from .swagger_client.models import AnalysisResultResponse
 from .config import config as thoth_config
@@ -71,6 +74,7 @@ from typing import Callable, Any, Union, Dict
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
 LAST_ANALYSIS_ID_FILE = ".thoth_last_analysis_id"
+_RECOMMENDATION_TYPES_URL = "https://thoth-station.ninja/recommendation-types/"
 
 _LOGGER = logging.getLogger(__name__)
 _RETRY_ON_ERROR_COUNT = int(os.getenv("THAMOS_RETRY_ON_ERROR_COUNT", 3))
@@ -270,13 +274,14 @@ def _retrieve_analysis_result(
             sleep(_RETRY_ON_ERROR_SLEEP)
 
 
-def _get_static_analysis(src_path: str = ".") -> typing.Optional[dict]:
+def get_static_analysis(src_path: str = ".") -> typing.Optional[dict]:
     """Get static analysis of files used in project."""
     # We are running in the root directory of project, use the root part for gathering static analysis.
     _LOGGER.info("Performing static analysis of sources to gather library usage")
     try:
         library_usage = gather_library_usage(
-            src_path, ignore_errors=True, without_standard_imports=True
+            src_path,
+            ignore_errors=True,
         )
     except FileNotFoundError:
         _LOGGER.warning("No library usage was aggregated - no Python sources found")
@@ -344,6 +349,7 @@ def advise_using_config(
     github_installation_id: typing.Optional[int] = None,
     github_base_repo_url: typing.Optional[str] = None,
     source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
+    labels: typing.Optional[Dict[str, str]] = None,
 ) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], bool]]:
     """Trigger advise, respecting the configuration file supplied directly as a string or as a file path."""
     try:
@@ -377,6 +383,7 @@ def advise_using_config(
         github_base_repo_url=github_base_repo_url,
         source_type=source_type,
         verify_tls=thoth_config.tls_verify,
+        labels=labels,
     )
 
 
@@ -410,6 +417,7 @@ def advise(
     stack_info: typing.Optional[Dict] = None,
     kebechet_metadata: typing.Optional[Dict] = None,
     verify_tls: bool = True,
+    labels: typing.Optional[Dict[str, str]] = None,
 ) -> typing.Optional[tuple]:
     """Submit a stack for adviser checks and wait for results."""
     if not pipfile:
@@ -438,6 +446,12 @@ def advise(
         )
         recommendation_type = next(filter(bool, priority))
 
+    _LOGGER.info(
+        "Using %r recommendation type - see %s",
+        recommendation_type,
+        _RECOMMENDATION_TYPES_URL,
+    )
+
     if no_user_stack and pipfile_lock:
         _LOGGER.warning(
             "The user stack found in the lock file will not be supplied as requested"
@@ -446,7 +460,7 @@ def advise(
 
     library_usage = None
     if not no_static_analysis:
-        library_usage = _get_static_analysis(src_path)
+        library_usage = get_static_analysis(src_path)
         _LOGGER.debug(
             "Library usage:%s",
             "\n" + json.dumps(library_usage, indent=2) if library_usage else None,
@@ -472,6 +486,10 @@ def advise(
         input_args["stack_info"] = stack_info
     if constraints:
         input_args["constraints"] = constraints
+    if labels:
+        input_args["labels"] = labels
+    elif runtime_environment.get("labels"):
+        input_args["labels"] = runtime_environment["labels"]
 
     advise_input = AdviseInput(**input_args)
     api_instance = AdviseApi(api_client)
@@ -571,90 +589,113 @@ def advise_here(
     stack_info: typing.Optional[Dict] = None,
     kebechet_metadata: typing.Optional[Dict] = None,
     verify_tls: bool = True,
+    labels: typing.Optional[Dict[str, str]] = None,
 ) -> typing.Optional[tuple]:
     """Run advise in current directory, requires no arguments."""
-    requirements_format = thoth_config.requirements_format
-    if requirements_format == "pipenv":
-        _LOGGER.info(
-            "Using Pipenv files to manage dependencies located in %r", os.getcwd()
-        )
-        pipfile_lock_exists = os.path.exists("Pipfile.lock")
-
-        if pipfile_lock_exists:
-            _LOGGER.info(
-                "Submitting Pipfile.lock as a base for user's stack scoring - see %s",
-                jl("user_stack"),
-            )
-
-        project = Project.from_files(
-            pipfile_path="Pipfile",
-            pipfile_lock_path="Pipfile.lock",
-            without_pipfile_lock=not os.path.exists("Pipfile.lock"),
-        )
-
-        if (
-            pipfile_lock_exists
-            and project.pipfile_lock.meta.hash["sha256"]
-            != project.pipfile.hash()["sha256"]
-        ):
-            _LOGGER.error(
-                "Pipfile hash stated in Pipfile.lock %r does not correspond to Pipfile hash %r - was Pipfile "
-                "adjusted? This error is not critical.",
-                project.pipfile_lock.meta.hash["sha256"][:6],
-                project.pipfile.hash()["sha256"][:6],
-            )
-    elif requirements_format in ("pip", "pip-tools", "pip-compile"):
-        _LOGGER.info(
-            "Using requirements.txt file to manage dependencies located in %r",
-            os.getcwd(),
-        )
-        project = Project.from_pip_compile_files(allow_without_lock=True)
-    else:
+    if runtime_environment_name is not None and runtime_environment is not None:
         raise ValueError(
-            f"Unknown configuration option for requirements format: {requirements_format!r}"
+            "Both runtime_environment and runtime_environment_name were set."
         )
 
-    constraints_str = None
-    if os.path.exists("constraints.txt"):
-        _LOGGER.info("Using constraints.txt file located in %r", os.getcwd())
-        with open("constraints.txt") as constraints_file:
-            constraints_str = constraints_file.read()
+    if runtime_environment is None and runtime_environment_name is None:
+        runtime_environment = thoth_config.get_runtime_environment()
 
-        # Try to load constraints before the request to verify their correctness without sending request to the backend.
-        Constraints.from_string(constraints_str)
+    if runtime_environment:
+        runtime_environment_name = runtime_environment["name"]
 
-    pipfile = project.pipfile.to_string()
-    pipfile_lock_str = project.pipfile_lock.to_string() if project.pipfile_lock else ""
-
-    return advise(
-        pipfile=pipfile,
-        pipfile_lock=pipfile_lock_str,
-        constraints=constraints_str,
-        recommendation_type=recommendation_type,
-        src_path=src_path,
-        runtime_environment=runtime_environment,
-        runtime_environment_name=runtime_environment_name,
-        dev=dev,
-        no_static_analysis=no_static_analysis,
-        no_user_stack=no_user_stack,
-        nowait=nowait,
-        force=force,
-        limit=limit,
-        count=count,
-        debug=debug,
-        origin=origin,
-        timeout=timeout,
-        source_type=source_type,
-        github_event_type=github_event_type,
-        github_check_run_id=github_check_run_id,
-        github_installation_id=github_installation_id,
-        github_base_repo_url=github_base_repo_url,
-        justification=justification,
-        stack_info=stack_info,
-        kebechet_metadata=kebechet_metadata,
-        verify_tls=verify_tls,
+    # get_overlays_directory() returns the current working directory if overlays_dir is not configured so this change
+    # should not result in a behavioral change
+    overlay_dir = thoth_config.get_overlays_directory(
+        runtime_environment_name=runtime_environment_name
     )
 
+    with cwd(overlay_dir):
+        requirements_format = thoth_config.requirements_format
+        if requirements_format == "pipenv":
+            _LOGGER.info(
+                "Using Pipenv files to manage dependencies located in %r", os.getcwd()
+            )
+            pipfile_lock_exists = os.path.exists("Pipfile.lock")
+
+            if pipfile_lock_exists:
+                _LOGGER.info(
+                    "Submitting Pipfile.lock as a base for user's stack scoring - see %s",
+                    jl("user_stack"),
+                )
+
+            project = Project.from_files(
+                pipfile_path="Pipfile",
+                pipfile_lock_path="Pipfile.lock",
+                without_pipfile_lock=not os.path.exists("Pipfile.lock"),
+            )
+
+            if (
+                pipfile_lock_exists
+                and project.pipfile_lock.meta.hash["sha256"]
+                != project.pipfile.hash()["sha256"]
+            ):
+                _LOGGER.error(
+                    "Pipfile hash stated in Pipfile.lock %r does not correspond to Pipfile hash %r - was Pipfile "
+                    "adjusted? This error is not critical.",
+                    project.pipfile_lock.meta.hash["sha256"][:6],
+                    project.pipfile.hash()["sha256"][:6],
+                )
+        elif requirements_format in ("pip", "pip-tools", "pip-compile"):
+            _LOGGER.info(
+                "Using requirements.txt file to manage dependencies located in %r",
+                os.getcwd(),
+            )
+            project = Project.from_pip_compile_files(allow_without_lock=True)
+        else:
+            raise ValueError(
+                f"Unknown configuration option for requirements format: {requirements_format!r}"
+            )
+
+        constraints_str = None
+        if os.path.exists("constraints.txt"):
+            _LOGGER.info("Using constraints.txt file located in %r", os.getcwd())
+            with open("constraints.txt") as constraints_file:
+                constraints_str = constraints_file.read()
+
+            # Try to load constraints before the request to verify their correctness without sending request to the
+            # backend.
+            Constraints.from_string(constraints_str)
+
+        pipfile = project.pipfile.to_string()
+        pipfile_lock_str = (
+            project.pipfile_lock.to_string() if project.pipfile_lock else ""
+        )
+
+        return advise(
+            pipfile=pipfile,
+            pipfile_lock=pipfile_lock_str,
+            constraints=constraints_str,
+            recommendation_type=recommendation_type,
+            src_path=src_path,
+            runtime_environment=runtime_environment,
+            runtime_environment_name=runtime_environment_name,
+            dev=dev,
+            no_static_analysis=no_static_analysis,
+            no_user_stack=no_user_stack,
+            nowait=nowait,
+            force=force,
+            limit=limit,
+            count=count,
+            debug=debug,
+            origin=origin,
+            timeout=timeout,
+            source_type=source_type,
+            github_event_type=github_event_type,
+            github_check_run_id=github_check_run_id,
+            github_installation_id=github_installation_id,
+            github_base_repo_url=github_base_repo_url,
+            justification=justification,
+            stack_info=stack_info,
+            kebechet_metadata=kebechet_metadata,
+            verify_tls=verify_tls,
+            labels=labels,
+        )
+
 
 @with_api_client
 def provenance_check(
@@ -945,7 +986,7 @@ def get_log(api_client: ApiClient, analysis_id: str = None):
 
 
 @with_api_client
-def get_status(api_client: ApiClient, analysis_id: str = None):
+def get_status(api_client: ApiClient, analysis_id: typing.Optional[str] = None):
     """Get status of an analysis - the analysis type and endpoint are automatically derived from analysis id.
 
     If analysis_id is not provided, its get from the last thamos call which stores it in a temporary file.
@@ -973,8 +1014,13 @@ def get_status(api_client: ApiClient, analysis_id: str = None):
 
 
 @with_api_client
-def get_analysis_results(api_client: ApiClient, analysis_id: str):
+def get_analysis_results(
+    api_client: ApiClient, analysis_id: typing.Optional[str] = None
+):
     """Get the analysis result from a given id."""
+    if not analysis_id:
+        analysis_id = _get_last_analysis_id()
+
     if analysis_id.startswith("package-extract-"):
         api_instance = ImageAnalysisApi(
             api_client
@@ -1037,7 +1083,7 @@ def install(
             if not os.path.isfile("Pipfile.lock"):
                 raise NoRequirementsFile(
                     f"No Pipfile.lock found in {os.getcwd()!r} needed to install dependencies, "
-                    "issue `thamos advise` resolve dpeendencies"
+                    "issue `thamos advise` resolve dependencies"
                 )
             if not os.path.isfile("Pipfile"):  # Required for computing digests.
                 raise NoRequirementsFile(
@@ -1113,19 +1159,21 @@ def install(
 
 
 @with_api_client
-def list_hardware_environments(api_client: ApiClient) -> typing.Dict[str, Any]:
-    """Get information about hardware for which Thoth can give recommendations."""
+def list_thoth_container_images(api_client: ApiClient) -> typing.Dict[str, Any]:
+    """Get available Thoth container images."""
     return (
-        HardwareApi(api_client)
-        .list_hardware_environments()
-        .to_dict()["hardware_environments"]
+        ContainerImagesApi(api_client)
+        .list_thoth_container_images()
+        .to_dict()["container_images"]
     )
 
 
 @with_api_client
-def list_thoth_s2i(api_client: ApiClient) -> typing.Dict[str, Any]:
-    """Get information about hardware for which Thoth can give recommendations."""
-    return S2iApi(api_client).list_s2i_python().to_dict()["s2i"]
+def list_python_environments(api_client: ApiClient) -> typing.Dict[str, Any]:
+    """Get available Python environments."""
+    return (
+        EnvironmentsApi(api_client).list_python_environments().to_dict()["environment"]
+    )
 
 
 @with_api_client
@@ -1136,6 +1184,76 @@ def list_python_package_indexes(api_client: ApiClient) -> typing.Dict[str, Any]:
     )
 
 
+@with_api_client
+def get_package_from_imported_packages(
+    api_client: ApiClient, import_name: str
+) -> typing.List[Dict[str, Any]]:
+    """Get all (package_name, package_version, index_url) triplets for given import package name."""
+    return (
+        PythonPackagesApi(api_client)
+        .get_package_from_imported_packages(import_name)
+        .to_dict()["package_names"]
+    )
+
+
+def get_verified_packages_from_static_analysis(src_path: str = "."):
+    """Get verified packages from invectio static analysis result."""
+    # 1. Obtain list of imports using invectio
+    result = get_static_analysis(src_path)
+    packages = []
+
+    if result:
+        packages = [p for p in result["report"].keys()]
+
+    # 2. For each import verify package (name, version, index) (whatprovides logic)
+    verified_packages = []
+
+    for import_name in packages:
+        unique_packages: typing.List[typing.Dict] = []
+
+        try:
+            imported_packages = get_package_from_imported_packages(import_name)
+
+            if imported_packages:
+                for package in imported_packages:
+                    if package["package_name"] not in [
+                        p["package_name"] for p in unique_packages
+                    ]:
+                        unique_packages.append(
+                            {
+                                "package_name": package["package_name"],
+                                "index_url": package["index_url"],
+                            }
+                        )
+                    else:
+                        existing_indexes = [
+                            p["index_url"]
+                            for p in unique_packages
+                            if p["package_name"] == package["package_name"]
+                        ]
+
+                        if package["index_url"] not in existing_indexes:
+                            unique_packages.append(
+                                {
+                                    "package_name": package["package_name"],
+                                    "index_url": package["index_url"],
+                                }
+                            )
+
+                for unique_package in unique_packages:
+                    verified_packages.append(unique_package)
+                    _LOGGER.info(
+                        f"Package name {unique_package['package_name']} identifed for import name {import_name}"
+                    )
+
+        except Exception as error:
+            _LOGGER.warning(
+                f"No packages identified for import name {import_name}: {error}"
+            )
+
+    return verified_packages
+
+
 def write_files(
     requirements: typing.Dict[str, Any],
     requirements_lock: typing.Dict[str, Any],
@@ -1169,7 +1287,8 @@ def load_files(requirements_format: str) -> typing.Tuple[str, typing.Optional[st
             )
 
         project = Project.from_files(
-            without_pipfile_lock=not os.path.exists("Pipfile.lock")
+            pipfile_path="Pipfile",
+            without_pipfile_lock=not os.path.exists("Pipfile.lock"),
         )
 
         if (
@@ -1240,3 +1359,209 @@ def write_configuration(
     _LOGGER.debug("Writing adjusted Thoth's configuration file")
     with open(".thoth.yaml", "w") as thoth_yaml_file:
         yaml.safe_dump(content, thoth_yaml_file)
+
+
+def collect_support_information_dict() -> Dict[str, Any]:
+    """Collect environment information suitable to report bugs or issues in a dictionary form."""
+    discovery = {}
+
+    for name in dir(thamos.discover):
+        if not name.startswith("discover_"):
+            continue
+
+        func = getattr(thamos.discover, name)
+        discovery[name[len("discover_") :]] = func()
+
+    thoth_config_content = None
+    try:
+        thoth_config_content = thoth_config.content
+    except Exception:
+        _LOGGER.exception("Failed to obtain content of Thoth's configuration file")
+
+    thoth_version = None
+    try:
+        thoth_version = thoth_config.get_thoth_version()
+    except Exception:
+        _LOGGER.exception("Failed to obtain Thoth version information")
+
+    pip_freeze = None
+    try:
+        pip_freeze = run_command("pip freeze").stdout.splitlines()
+    except Exception:
+        _LOGGER.exception("Failed to obtain packages present in the environment")
+
+    last_analysis_id = None
+    try:
+        last_analysis_id = _get_last_analysis_id()
+    except FileNotFoundError:
+        _LOGGER.warning("Cannot retrieve last analysis identifier")
+        pass
+
+    result = {
+        "environment": {
+            k: v for k, v in os.environ.items() if k.startswith(("THAMOS_", "THOTH_"))
+        },
+        "is_s2i": _is_s2i(),
+        "last_analysis_id": last_analysis_id,
+        "thoth_config": thoth_config_content,
+        "thoth_version": thoth_version,
+        "thamos_version": thamos_version,
+        "os_name": os.name,
+        "platform_machine": platform.machine(),
+        "platform_python_implementation": platform.python_implementation(),
+        "platform_release": platform.release(),
+        "platform_system": platform.system(),
+        "platform_version": platform.version(),
+        "python_full_version": platform.python_version(),
+        "python_version": platform.python_version()[:3],
+        "sys_platform": sys.platform,
+        "pip_freeze": pip_freeze,
+        "discovery": discovery,
+    }
+    return result
+
+
+def _get_package_entry_str(
+    pipfile_lock: Dict[str, Dict[str, typing.Any]], package_name: str
+) -> str:
+    """Get information about the given package so that they are printed to users."""
+    package_entry = pipfile_lock["default"].get(
+        package_name, pipfile_lock["develop"].get(package_name)
+    )
+    if not package_entry:
+        # Any error spotted?
+        return "UNKNOWN==UNKNOWN from UNKNOWN"
+
+    index_url = "UNKNOWN"
+    for source in pipfile_lock["_meta"]["sources"]:
+        if package_entry.get("index") == source["name"]:
+            index_url = source["url"]
+            break
+
+    return f"{package_name}{package_entry.get('version')} from {index_url}"
+
+
+def _traverse_edges(
+    pipfile_lock: Dict[str, Dict[str, str]],
+    nodes: typing.List[str],
+    edges: typing.List[typing.List[int]],
+    seen_nodes: typing.Set[int],
+    *,
+    starting_node: int,
+    indent: int = 2,
+    fold: bool = True,
+) -> None:
+    """Traverse edges of the dependency graph and show its structure to terminal."""
+    for edge in edges:
+        if edge[0] == starting_node:
+            print(
+                " " * indent
+                + f"↳ {_get_package_entry_str(pipfile_lock, nodes[edge[1]])}"
+            )
+            if edge[1] not in seen_nodes:
+                if fold:
+                    seen_nodes.add(edge[1])
+
+                _traverse_edges(
+                    pipfile_lock,
+                    nodes,
+                    edges,
+                    seen_nodes=seen_nodes | {edge[1]},
+                    starting_node=edge[1],
+                    indent=indent + 2,
+                    fold=fold,
+                )
+            else:
+                # Note cyclic dependency.
+                print(" " * (indent + 2) + "↳ …")
+
+
+def print_dependency_graph_from_adviser_document(
+    adviser_document: typing.Dict[str, typing.Any], *, fold: bool = True
+) -> bool:
+    """Print dependency graph to stdout from the given document."""
+    if not adviser_document["report"]["products"]:
+        _LOGGER.error(
+            "No dependency graph to show, was the adviser request successful?"
+        )
+        return False
+
+    if not adviser_document["report"]["products"][0].get("dependency_graph"):
+        # XXX: we can remove this check over time.
+        _LOGGER.error(
+            "No dependency graph to show, was the result computed with recent adviser?"
+        )
+        return False
+
+    nodes = adviser_document["report"]["products"][0]["dependency_graph"]["nodes"]
+    edges = adviser_document["report"]["products"][0]["dependency_graph"]["edges"]
+
+    nodes_idx = {n: i for i, n in enumerate(nodes)}
+    # Keep a list of flags stating the given package is a root dependency.
+    pipfile = adviser_document["report"]["products"][0]["project"]["requirements"]
+    pipfile_lock = adviser_document["report"]["products"][0]["project"][
+        "requirements_locked"
+    ]
+    for direct_dependency in chain(
+        pipfile["packages"].keys(), pipfile["dev-packages"].keys()
+    ):
+        direct_dependency_idx = nodes_idx[direct_dependency]
+        print(f"→ {_get_package_entry_str(pipfile_lock, direct_dependency)}")
+        _traverse_edges(
+            pipfile_lock,
+            nodes,
+            edges,
+            seen_nodes={direct_dependency_idx},
+            starting_node=direct_dependency_idx,
+            fold=fold,
+        )
+
+    return True
+
+
+def print_dependency_graph(
+    analysis_id: typing.Optional[str] = None, *, fold: bool = True
+) -> bool:
+    """Print dependency graph to stdout produced by the given adviser."""
+    if not analysis_id:
+        adviser_document, error = get_analysis_results()
+    else:
+        adviser_document, error = get_analysis_results(analysis_id)
+
+    if error:
+        print("Cannot print dependency graph as advise was not successful")
+        return False
+
+    return print_dependency_graph_from_adviser_document(adviser_document, fold=fold)
+
+
+def add_requirements_to_project(
+    requirement: typing.List[str],
+    runtime_environment: typing.Optional[str],
+    index_url: str,
+    dev: bool,
+):
+    """Add requirements to project."""
+    project = thoth_config.get_project(runtime_environment, missing_dir_ok=True)
+
+    for req in requirement:
+        _LOGGER.info(
+            "Adding %r to %s requirements of runtime environment %r",
+            req,
+            "development" if dev else "default",
+            project.runtime_environment.name,
+        )
+        project.pipfile.add_requirement(
+            req, is_dev=dev, index_url=index_url, force=True
+        )
+        runtime_environment_config = thoth_config.get_runtime_environment(
+            runtime_environment
+        )
+        python_version = runtime_environment_config.get("python_version")
+        if python_version:
+            project.set_python_version(python_version)
+
+    _LOGGER.warning(
+        "Changes done might require triggering new advise to resolve dependencies"
+    )
+    thoth_config.save_project(project)
