@@ -7,10 +7,8 @@ from thamos.swagger_client.thoth.advise_api import AdviseApi
 from thamos.swagger_client.thoth.build_analysis_api import BuildAnalysisApi
 from thamos.swagger_client.thoth.buildlogs_api import BuildlogsApi
 from thamos.swagger_client.thoth.container_images_api import ContainerImagesApi
-from thamos.swagger_client.thoth.git_hub_api import GitHubApi
-from thamos.swagger_client.thoth.hardware_api import HardwareApi
+from thamos.swagger_client.thoth.environments_api import EnvironmentsApi
 from thamos.swagger_client.thoth.image_analysis_api import ImageAnalysisApi
 from thamos.swagger_client.thoth.kebechet_api import KebechetApi
 from thamos.swagger_client.thoth.provenance_api import ProvenanceApi
 from thamos.swagger_client.thoth.python_packages_api import PythonPackagesApi
-from thamos.swagger_client.thoth.s2i_api import S2iApi
